'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

var _debug = require('debug');

var _debug2 = _interopRequireDefault(_debug);

var _cloneDeep = require('lodash/cloneDeep');

var _cloneDeep2 = _interopRequireDefault(_cloneDeep);

var _pImmediate = require('p-immediate');

var _pImmediate2 = _interopRequireDefault(_pImmediate);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _asyncToGenerator(fn) { return function () { var gen = fn.apply(this, arguments); return new Promise(function (resolve, reject) { function step(key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { return Promise.resolve(value).then(function (value) { step("next", value); }, function (err) { step("throw", err); }); } } return step("next"); }); }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

/**
 * Debug.
 *
 * @type {Function}
 */

var debug = (0, _debug2.default)('nexus:store-memory');

/**
 * The default, memory-only store adapter.
 *
 * @type {MemoryStore}
 */

var MemoryStore = function () {

  /**
   * Constructor.
   *
   * @param {Object} config
   */

  function MemoryStore() {
    var config = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};

    _classCallCheck(this, MemoryStore);

    this.snapshots = config.snapshots || {};
    this.operations = config.operations || {};
  }

  /**
   * Save an `operation` and `snapshot` on a document by `id`.
   *
   * @param {String} id
   * @param {Object} operation
   * @param {Object} snapshot
   * @param {Object} options
   * @return {Boolean}
   */

  _createClass(MemoryStore, [{
    key: 'commit',
    value: function () {
      var _ref = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee(id, operation, snapshot, options) {
        var ops, version;
        return regeneratorRuntime.wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                debug('commit', { id: id, operation: operation, snapshot: snapshot, options: options });

                _context.next = 3;
                return (0, _pImmediate2.default)();

              case 3:
                ops = this.operations[id] || (this.operations[id] = []);
                version = ops && ops.length || 0;

                if (!(snapshot.version !== version + 1)) {
                  _context.next = 7;
                  break;
                }

                return _context.abrupt('return', false);

              case 7:

                ops[operation.version] = (0, _cloneDeep2.default)(operation);
                this.snapshots[id] = (0, _cloneDeep2.default)(snapshot);
                return _context.abrupt('return', true);

              case 10:
              case 'end':
                return _context.stop();
            }
          }
        }, _callee, this);
      }));

      function commit(_x2, _x3, _x4, _x5) {
        return _ref.apply(this, arguments);
      }

      return commit;
    }()

    /**
     * Find the latest snapshot of a document by `id`.
     *
     * @param {String} id
     * @param {Object} options
     * @return {Object}
     */

  }, {
    key: 'findSnapshot',
    value: function () {
      var _ref2 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee2(id, options) {
        var snapshot, ret;
        return regeneratorRuntime.wrap(function _callee2$(_context2) {
          while (1) {
            switch (_context2.prev = _context2.next) {
              case 0:
                debug('findSnapshot', { id: id, options: options });

                _context2.next = 3;
                return (0, _pImmediate2.default)();

              case 3:
                snapshot = this.snapshots[id];

                if (snapshot) {
                  _context2.next = 6;
                  break;
                }

                return _context2.abrupt('return');

              case 6:
                ret = (0, _cloneDeep2.default)(snapshot);
                return _context2.abrupt('return', ret);

              case 8:
              case 'end':
                return _context2.stop();
            }
          }
        }, _callee2, this);
      }));

      function findSnapshot(_x6, _x7) {
        return _ref2.apply(this, arguments);
      }

      return findSnapshot;
    }()

    /**
     * List the operations for a document between versions `from` and `to`. The
     * operations should include `from` and include everything up to, but NOT
     * including `to`.
     *
     * @param {String} id
     * @param {Number} from
     * @param {Number} to
     * @param {Object} options
     * @return {Array}
     */

  }, {
    key: 'listOperations',
    value: function () {
      var _ref3 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee3(id, from, to, options) {
        var ops, slice, ret;
        return regeneratorRuntime.wrap(function _callee3$(_context3) {
          while (1) {
            switch (_context3.prev = _context3.next) {
              case 0:
                debug('listOperations', { id: id, from: from, to: to, options: options });

                _context3.next = 3;
                return (0, _pImmediate2.default)();

              case 3:
                ops = this.operations[id];

                if (ops) {
                  _context3.next = 6;
                  break;
                }

                return _context3.abrupt('return', []);

              case 6:
                slice = ops.slice(from, to);
                ret = (0, _cloneDeep2.default)(slice);
                return _context3.abrupt('return', ret);

              case 9:
              case 'end':
                return _context3.stop();
            }
          }
        }, _callee3, this);
      }));

      function listOperations(_x8, _x9, _x10, _x11) {
        return _ref3.apply(this, arguments);
      }

      return listOperations;
    }()
  }]);

  return MemoryStore;
}();

/**
 * Export.
 *
 * @type {MemoryStore}
 */

exports.default = MemoryStore;