'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

var _debug = require('debug');

var _debug2 = _interopRequireDefault(_debug);

var _events = require('events');

var _events2 = _interopRequireDefault(_events);

var _document = require('./document');

var _document2 = _interopRequireDefault(_document);

var _checkers = require('../utils/checkers');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _asyncToGenerator(fn) { return function () { var gen = fn.apply(this, arguments); return new Promise(function (resolve, reject) { function step(key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { return Promise.resolve(value).then(function (value) { step("next", value); }, function (err) { step("throw", err); }); } } return step("next"); }); }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _possibleConstructorReturn(self, call) { if (!self) { throw new ReferenceError("this hasn't been initialised - super() hasn't been called"); } return call && (typeof call === "object" || typeof call === "function") ? call : self; }

function _inherits(subClass, superClass) { if (typeof superClass !== "function" && superClass !== null) { throw new TypeError("Super expression must either be null or a function, not " + typeof superClass); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, enumerable: false, writable: true, configurable: true } }); if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass; }

/**
 * Debug.
 *
 * @type {Function}
 */

var debug = (0, _debug2.default)('nexus:client');

/**
 * Client.
 *
 * @type {Client}
 */

var Client = function (_Emitter) {
  _inherits(Client, _Emitter);

  /**
   * Constructor.
   *
   * @param {Object} config
   */

  function Client() {
    var _this2 = this;

    var config = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};

    _classCallCheck(this, Client);

    var _this = _possibleConstructorReturn(this, (Client.__proto__ || Object.getPrototypeOf(Client)).call(this));

    _this.onTransportOpen = function () {
      _this.send({ action: 'connect' });
    };

    _this.onTransportClose = function (reason) {
      _this.sequence = 1;
      _this.id = null;
      _this.connected = false;
      _this.emit('disconnect', reason);
    };

    _this.onTransportReadable = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee() {
      var message, doc;
      return regeneratorRuntime.wrap(function _callee$(_context) {
        while (1) {
          switch (_context.prev = _context.next) {
            case 0:
              message = _this.transport.read();

              (0, _checkers.checkMessage)(message);

              debug('onMessage', { message: message });

              if (!(message.action === 'connect')) {
                _context.next = 8;
                break;
              }

              _this.id = message.id;
              _this.connected = true;
              _this.emit('connect');
              return _context.abrupt('return');

            case 8:
              doc = _this.documents[message.document];

              if (doc) {
                _context.next = 11;
                break;
              }

              return _context.abrupt('return');

            case 11:
              _context.prev = 11;
              _context.next = 14;
              return doc.onMessage(message);

            case 14:
              _context.next = 20;
              break;

            case 16:
              _context.prev = 16;
              _context.t0 = _context['catch'](11);

              _this.emit('error', _context.t0);
              return _context.abrupt('return');

            case 20:
            case 'end':
              return _context.stop();
          }
        }
      }, _callee, _this2, [[11, 16]]);
    }));

    _this.onTransportError = function (error) {
      _this.emit('error', error);
    };

    var transport = config.transport,
        types = config.types;

    (0, _checkers.checkTransport)(transport);
    (0, _checkers.checkTypes)(types);

    _this.transport = transport;
    _this.types = types;
    _this.documents = {};
    _this.id = null;
    _this.sequence = 1;
    _this.connected = false;

    transport.on('open', _this.onTransportOpen);
    transport.on('close', _this.onTransportClose);
    transport.on('readable', _this.onTransportReadable);
    transport.on('error', _this.onTransportError);

    if (!transport.closed) {
      _this.onTransportOpen();
    }
    return _this;
  }

  /**
   * Open a new document by `id`.
   *
   * @param {String} id
   * @return {Document}
   */

  _createClass(Client, [{
    key: 'open',
    value: function open(id) {
      var _this3 = this;

      debug('open', id);

      var document = this.documents[id];
      if (document) return document;

      var types = this.types;

      document = new _document2.default({ id: id, types: types, client: this });
      this.documents[id] = document;

      document.once('close', function () {
        delete _this3.documents[id];
      });

      return document;
    }

    /**
     * Wait for all of the client's document's inflight pending to complete.
     */

  }, {
    key: 'settle',
    value: function () {
      var _ref2 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee2() {
        var promises;
        return regeneratorRuntime.wrap(function _callee2$(_context2) {
          while (1) {
            switch (_context2.prev = _context2.next) {
              case 0:
                promises = Object.values(this.documents).map(function (d) {
                  return d.settle();
                });
                _context2.next = 3;
                return Promise.all(promises);

              case 3:
              case 'end':
                return _context2.stop();
            }
          }
        }, _callee2, this);
      }));

      function settle() {
        return _ref2.apply(this, arguments);
      }

      return settle;
    }()

    /**
     * Send a `message` to the server.
     *
     * @param {Object} message
     */

  }, {
    key: 'send',
    value: function send(message) {
      var _this4 = this;

      if (!this.connected && message.action !== 'connect') {
        return this.once('connect', function () {
          _this4.send(message);
        });
      }

      debug('send', message);

      (0, _checkers.checkMessage)(message);
      this.transport.write(message);
    }

    /**
     * On transport open.
     */

    /**
     * On transport close.
     *
     * @param {String} reason
     */

    /**
     * On transport readable.
     */

    /**
     * On transport error.
     *
     * @param {Error} error
     */

  }]);

  return Client;
}(_events2.default);

/**
 * Export.
 *
 * @type {Client}
 */

exports.default = Client;