'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

var _debug = require('debug');

var _debug2 = _interopRequireDefault(_debug);

var _events = require('events');

var _events2 = _interopRequireDefault(_events);

var _pDefer = require('p-defer');

var _pDefer2 = _interopRequireDefault(_pDefer);

var _nexusError = require('../utils/nexus-error');

var _nexusError2 = _interopRequireDefault(_nexusError);

var _checkers = require('../utils/checkers');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _asyncToGenerator(fn) { return function () { var gen = fn.apply(this, arguments); return new Promise(function (resolve, reject) { function step(key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { return Promise.resolve(value).then(function (value) { step("next", value); }, function (err) { step("throw", err); }); } } return step("next"); }); }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _possibleConstructorReturn(self, call) { if (!self) { throw new ReferenceError("this hasn't been initialised - super() hasn't been called"); } return call && (typeof call === "object" || typeof call === "function") ? call : self; }

function _inherits(subClass, superClass) { if (typeof superClass !== "function" && superClass !== null) { throw new TypeError("Super expression must either be null or a function, not " + typeof superClass); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, enumerable: false, writable: true, configurable: true } }); if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass; }

/**
 * Debug.
 *
 * @type {Function}
 */

var debug = (0, _debug2.default)('nexus:document');

/**
 * Document.
 *
 * @type {Document}
 */

var Document = function (_Emitter) {
  _inherits(Document, _Emitter);

  /**
   * Constructor.
   *
   * @param {Object} config
   */

  function Document() {
    var config = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};

    _classCallCheck(this, Document);

    var _this = _possibleConstructorReturn(this, (Document.__proto__ || Object.getPrototypeOf(Document)).call(this));

    _initialiseProps.call(_this);

    var id = config.id,
        client = config.client,
        types = config.types;

    (0, _checkers.checkTypes)(types);

    _this.id = id;
    _this.version = null;
    _this.type = null;
    _this.value = undefined;

    _this.types = types;
    _this.deferreds = {};
    _this.queue = [];
    _this.inflight = null;
    _this.closed = false;
    _this.subscribed = false;

    _this.client = client;
    _this.client.on('disconnect', _this.onClientDisconnect);
    return _this;
  }

  /**
   * Close the document.
   */

  _createClass(Document, [{
    key: 'close',
    value: function () {
      var _ref = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee() {
        return regeneratorRuntime.wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                debug('close', this.id);

                _context.next = 3;
                return this.settle();

              case 3:
                _context.next = 5;
                return this.unsubscribe();

              case 5:

                this.client.removeListener('disconnect', this.onClientDisconnect);
                this.closed = true;
                this.emit('close');

              case 8:
              case 'end':
                return _context.stop();
            }
          }
        }, _callee, this);
      }));

      function close() {
        return _ref.apply(this, arguments);
      }

      return close;
    }()

    /**
     * Wait for the document's pending messages to complete.
     */

  }, {
    key: 'settle',
    value: function () {
      var _ref2 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee2() {
        var _deferreds, fetch, subscribe, unsubscribe, queue;

        return regeneratorRuntime.wrap(function _callee2$(_context2) {
          while (1) {
            switch (_context2.prev = _context2.next) {
              case 0:
                _deferreds = this.deferreds, fetch = _deferreds.fetch, subscribe = _deferreds.subscribe, unsubscribe = _deferreds.unsubscribe, queue = _deferreds.queue;
                _context2.next = 3;
                return Promise.all([fetch && fetch.promise, subscribe && subscribe.promise, unsubscribe && unsubscribe.promise, queue && queue.promise]);

              case 3:
              case 'end':
                return _context2.stop();
            }
          }
        }, _callee2, this);
      }));

      function settle() {
        return _ref2.apply(this, arguments);
      }

      return settle;
    }()

    /**
     * Fetch the document from the server. Unlike subscribe, this is only a
     * one-time fetch that won't listen for subsequent operation messages.
     *
     * @param {Object} options
     * @return {Promise}
     */

  }, {
    key: 'fetch',
    value: function fetch() {
      var options = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};

      debug('fetch', this.id);

      var _options$force = options.force,
          force = _options$force === undefined ? false : _options$force;


      if (!this.deferreds.fetch || force) {
        this.send({
          action: 'fetch',
          document: this.id,
          version: this.version
        });
      }

      var deferred = this.createDeferred('fetch');
      return deferred.promise;
    }

    /**
     * Subscribe to a document, fetching it from the server and listening for any
     * subsequent operations that are submitted by other connected clients.
     *
     * @param {Object} options
     * @return {Promise}
     */

  }, {
    key: 'subscribe',
    value: function subscribe() {
      var _this2 = this;

      var options = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};

      debug('subscribe', this.id);

      var _options$force2 = options.force,
          force = _options$force2 === undefined ? false : _options$force2;


      if (!this.deferreds.subscribe || force) {
        this.send({
          action: 'subscribe',
          document: this.id,
          version: this.version
        });
      }

      var deferred = this.createDeferred('subscribe');

      // Update the `subscribed` flag when the deferred resolves, so that we can
      // know whether to try re-subscribing on reconnect.
      deferred.promise.then(function (value) {
        _this2.subscribed = true;
        return value;
      });

      return deferred.promise;
    }

    /**
     * Unsubscribe from the document, which stops listening for new operations.
     *
     * @param {Object} options
     * @return {Promise}
     */

  }, {
    key: 'unsubscribe',
    value: function unsubscribe() {
      var options = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};

      debug('unsubscribe', this.id);

      var _options$force3 = options.force,
          force = _options$force3 === undefined ? false : _options$force3;

      // Update the `subscribed` flag optimistically, in case the message errors.

      this.subscribed = false;

      if (!this.deferreds.unsubscribe || force) {
        this.send({
          action: 'unsubscribe',
          document: this.id
        });
      }

      var deferred = this.createDeferred('unsubscribe');
      return deferred.promise;
    }

    /**
     * Send a `message` to the server.
     *
     * @param {Object} message
     */

  }, {
    key: 'send',
    value: function send(message) {
      if (this.closed) {
        throw new Error('Tried to send a message for document "' + this.id + '", but it was already closed.');
      }

      debug('send', this.id, { message: message });
      this.client.send(message);
    }

    /**
     * Submit an `operation` to the server by adding it to the operation queue.
     *
     * @param {Object} operation
     */

  }, {
    key: 'submit',
    value: function submit(operation) {
      var _this3 = this;

      var type = this.type,
          version = this.version,
          id = this.id,
          queue = this.queue,
          value = this.value;


      if (type == null || version == null) {
        throw new Error('You must call `document.fetch()` or `document.subscribe()` before you can submit operations to a document.');
      }

      debug('submit', id, { operation: operation });

      if (type.normalize) {
        operation = type.normalize(operation);
      }

      var last = queue[queue.length - 1];
      var message = {
        action: 'submit',
        document: id,
        data: operation

        // If the OT type defines a `compose` method, it must be able to compose any
        // operation into another, so we can compose into the last queued instead.
      };if (type.compose && last) {
        last.data = type.compose(last.data, message.data);
      }

      // Otherwise, just add the message to the queue.
      else {
          queue.push(message);
          this.createDeferred('queue');
        }

      // Regardless, apply the non-composed message, since even if we're composing
      // the last was already applied individually.
      this.value = type.apply(value, message.data);

      // Delay flushing to allow composing multiple queued operations.
      process.nextTick(function () {
        return _this3.flush();
      });
    }

    /**
     * Flush any queued operations to the server.
     *
     * @param {Object} options
     */

  }, {
    key: 'flush',
    value: function flush() {
      var options = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
      var _options$force4 = options.force,
          force = _options$force4 === undefined ? false : _options$force4;


      if (!this.client.connected) return;
      if (this.deferreds.submit && !force) return;

      var message = this.inflight || this.queue.shift();

      if (!message) {
        this.resolveDeferred('queue');
        return;
      }

      debug('flush', this.id, { message: message });

      // Since the message could have been an inflight one that got rejected, we
      // need to set the version right before flushing, not when queueing.
      message.version = this.version;

      // Do not overwrite an existing `id`, in case this is a retry of an
      // existing operations or the client has reconnected and gotten a new ID.
      if (message.id == null) {
        message.id = this.client.id + ':' + this.client.sequence++;
      }

      this.inflight = message;
      this.createDeferred('submit');
      this.send(message);
    }

    /**
     * Update the document's version and value with a `snapshot`.
     *
     * @param {Object} snapshot
     */

  }, {
    key: 'update',
    value: function () {
      var _ref3 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee3(snapshot) {
        var data, version, type;
        return regeneratorRuntime.wrap(function _callee3$(_context3) {
          while (1) {
            switch (_context3.prev = _context3.next) {
              case 0:
                data = snapshot.data, version = snapshot.version;


                (0, _checkers.checkSnapshot)(snapshot);
                (0, _checkers.checkType)(snapshot.type, this.types);

                type = this.types.find(function (t) {
                  return t.name === snapshot.type;
                });

                // If we've already ingested the initial snapshot, we can't do it again. We
                // need to fetch operations to transform and apply to the existing snapshot.
                // Although if the version is old, we can just ignore it.

                if (!this.type) {
                  _context3.next = 9;
                  break;
                }

                if (!(version > this.version)) {
                  _context3.next = 8;
                  break;
                }

                _context3.next = 8;
                return this.fetch();

              case 8:
                return _context3.abrupt('return');

              case 9:

                this.version = version;
                this.type = type;
                this.value = type.deserialize ? type.deserialize(data) : data;

              case 12:
              case 'end':
                return _context3.stop();
            }
          }
        }, _callee3, this);
      }));

      function update(_x6) {
        return _ref3.apply(this, arguments);
      }

      return update;
    }()

    /**
     * Create a deferred object with `action`.
     *
     * @param {String} action
     */

  }, {
    key: 'createDeferred',
    value: function createDeferred(action) {
      var deferred = this.deferreds[action];
      if (deferred) return deferred;

      debug('createDeferred', { action: action });

      deferred = this.deferreds[action] = (0, _pDefer2.default)();
      return deferred;
    }

    /**
     * Resolve a deferred object by `action` with `value`.
     *
     * @param {String} action
     * @param {Mixed} value
     */

  }, {
    key: 'resolveDeferred',
    value: function resolveDeferred(action, value) {
      var deferred = this.deferreds[action];
      if (!deferred) return;

      debug('resolveDeferred', { action: action, value: value });

      deferred.resolve(value);
      delete this.deferreds[action];
    }

    /**
     * Reject a deferred object by `action` with `err`.
     *
     * @param {String} action
     * @param {Mixed} err
     */

  }, {
    key: 'rejectDeferred',
    value: function rejectDeferred(action, err) {
      var deferred = this.deferreds[action];
      if (!deferred) return;

      debug('rejectDeferred', { action: action, err: err });

      deferred.reject(err);
      delete this.deferreds[action];
    }

    /**
     * On message.
     *
     * @param {Object} message
     */

  }, {
    key: 'onMessage',
    value: function () {
      var _ref4 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee4(message) {
        var err, action, error;
        return regeneratorRuntime.wrap(function _callee4$(_context4) {
          while (1) {
            switch (_context4.prev = _context4.next) {
              case 0:
                debug('onMessage', this.id, { message: message });

                (0, _checkers.checkMessage)(message);

                err = null;
                action = message.action, error = message.error;


                if (error) {
                  err = new _nexusError2.default('server_error', { error: error });
                  err.code = error.code;
                  err.data = message;
                }

                _context4.t0 = action;
                _context4.next = _context4.t0 === 'broadcast' ? 8 : _context4.t0 === 'fetch' ? 11 : _context4.t0 === 'submit' ? 14 : _context4.t0 === 'subscribe' ? 17 : _context4.t0 === 'unsubscribe' ? 20 : 23;
                break;

              case 8:
                _context4.next = 10;
                return this.onBroadcast(err, message);

              case 10:
                return _context4.abrupt('return', _context4.sent);

              case 11:
                _context4.next = 13;
                return this.onFetch(err, message);

              case 13:
                return _context4.abrupt('return', _context4.sent);

              case 14:
                _context4.next = 16;
                return this.onSubmit(err, message);

              case 16:
                return _context4.abrupt('return', _context4.sent);

              case 17:
                _context4.next = 19;
                return this.onSubscribe(err, message);

              case 19:
                return _context4.abrupt('return', _context4.sent);

              case 20:
                _context4.next = 22;
                return this.onUnsubscribe(err, message);

              case 22:
                return _context4.abrupt('return', _context4.sent);

              case 23:
              case 'end':
                return _context4.stop();
            }
          }
        }, _callee4, this);
      }));

      function onMessage(_x7) {
        return _ref4.apply(this, arguments);
      }

      return onMessage;
    }()

    /**
     * On "broadcast" message.
     *
     * @param {Object} err
     * @param {Object} message
     */

  }, {
    key: 'onBroadcast',
    value: function () {
      var _ref5 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee5(err, message) {
        var inflight, type, value, queue, version, data, locals, _iteratorNormalCompletion, _didIteratorError, _iteratorError, _iterator, _step, local, l, s;

        return regeneratorRuntime.wrap(function _callee5$(_context5) {
          while (1) {
            switch (_context5.prev = _context5.next) {
              case 0:
                inflight = this.inflight, type = this.type, value = this.value, queue = this.queue;
                version = message.version, data = message.data;

                // If the document isn't yet fetched, the version is unset, so fetch it.

                if (!(this.version == null)) {
                  _context5.next = 6;
                  break;
                }

                _context5.next = 5;
                return this.fetch();

              case 5:
                return _context5.abrupt('return');

              case 6:
                if (!(version > this.version)) {
                  _context5.next = 10;
                  break;
                }

                _context5.next = 9;
                return this.fetch();

              case 9:
                return _context5.abrupt('return');

              case 10:
                if (!(version < this.version)) {
                  _context5.next = 12;
                  break;
                }

                return _context5.abrupt('return');

              case 12:

                // Otherwise, this is a new operation from another connected client. So
                // transform local operations with it, increment and apply it.
                locals = queue.slice();

                if (inflight) locals.unshift(inflight);

                _iteratorNormalCompletion = true;
                _didIteratorError = false;
                _iteratorError = undefined;
                _context5.prev = 17;
                for (_iterator = locals[Symbol.iterator](); !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
                  local = _step.value;
                  l = type.transform(local.data, message.data, 'left');
                  s = type.transform(message.data, local.data, 'right');

                  local.data = l;
                  message.data = s;
                }

                _context5.next = 25;
                break;

              case 21:
                _context5.prev = 21;
                _context5.t0 = _context5['catch'](17);
                _didIteratorError = true;
                _iteratorError = _context5.t0;

              case 25:
                _context5.prev = 25;
                _context5.prev = 26;

                if (!_iteratorNormalCompletion && _iterator.return) {
                  _iterator.return();
                }

              case 28:
                _context5.prev = 28;

                if (!_didIteratorError) {
                  _context5.next = 31;
                  break;
                }

                throw _iteratorError;

              case 31:
                return _context5.finish(28);

              case 32:
                return _context5.finish(25);

              case 33:
                this.value = type.apply(value, message.data);
                this.version++;
                this.emit('operation', data, { source: 'remote' });

              case 36:
              case 'end':
                return _context5.stop();
            }
          }
        }, _callee5, this, [[17, 21, 25, 33], [26,, 28, 32]]);
      }));

      function onBroadcast(_x8, _x9) {
        return _ref5.apply(this, arguments);
      }

      return onBroadcast;
    }()

    /**
     * On "fetch" message.
     *
     * @param {Object} err
     * @param {Object} message
     */

  }, {
    key: 'onFetch',
    value: function () {
      var _ref6 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee6(err, message) {
        var snapshot;
        return regeneratorRuntime.wrap(function _callee6$(_context6) {
          while (1) {
            switch (_context6.prev = _context6.next) {
              case 0:
                if (!err) {
                  _context6.next = 2;
                  break;
                }

                return _context6.abrupt('return', this.rejectDeferred('fetch', err));

              case 2:
                snapshot = message.snapshot;

                if (!snapshot) {
                  _context6.next = 6;
                  break;
                }

                _context6.next = 6;
                return this.update(snapshot);

              case 6:
                this.resolveDeferred('fetch', this.value);
                if (snapshot) this.emit('fetch', this.value);

              case 8:
              case 'end':
                return _context6.stop();
            }
          }
        }, _callee6, this);
      }));

      function onFetch(_x10, _x11) {
        return _ref6.apply(this, arguments);
      }

      return onFetch;
    }()

    /**
     * On "submit" message.
     *
     * @param {Object} err
     * @param {Object} message
     */

  }, {
    key: 'onSubmit',
    value: function () {
      var _ref7 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee7(err, message) {
        var inflight, id, version, data, e, _e, _e2;

        return regeneratorRuntime.wrap(function _callee7$(_context7) {
          while (1) {
            switch (_context7.prev = _context7.next) {
              case 0:
                inflight = this.inflight;
                id = message.id, version = message.version, data = message.data;

                // If there was an error, the server rejected it. To get back into a normal
                // state, we need to re-fetch the document.

                if (!err) {
                  _context7.next = 11;
                  break;
                }

                this.type = null;
                this.version = null;
                this.inflight = null;
                this.queue = [];
                _context7.next = 9;
                return this.fetch();

              case 9:

                this.rejectDeferred('submit', err);
                return _context7.abrupt('return');

              case 11:
                if (inflight) {
                  _context7.next = 15;
                  break;
                }

                e = new Error('Unexpectedly received acknowledgement for an operation when none was in flight.');

                this.rejectDeferred('submit', e);
                return _context7.abrupt('return');

              case 15:
                if (!(id !== inflight.id)) {
                  _context7.next = 19;
                  break;
                }

                _e = new Error('Unexpectedly received acknowledgement for an operation with an ID that did not match the current in flight operation\'s ID. Expected "' + inflight.id + '", but received "' + id + '".');

                this.rejectDeferred('submit', _e);
                return _context7.abrupt('return');

              case 19:
                if (!(version > this.version)) {
                  _context7.next = 23;
                  break;
                }

                _e2 = new Error('Unexpectedly received acknowledgement for an operation with a version greater than the document\'s current version. Expected: "' + this.version + '", but received "' + version + '".');

                this.rejectDeferred('submit', _e2);
                return _context7.abrupt('return');

              case 23:
                if (!(version < this.version)) {
                  _context7.next = 28;
                  break;
                }

                _context7.next = 26;
                return this.fetch();

              case 26:
                this.resolveDeferred('submit');
                return _context7.abrupt('return');

              case 28:

                this.version++;
                this.inflight = null;
                this.resolveDeferred('submit');
                this.flush();
                this.emit('operation', data, { source: 'local' });

              case 33:
              case 'end':
                return _context7.stop();
            }
          }
        }, _callee7, this);
      }));

      function onSubmit(_x12, _x13) {
        return _ref7.apply(this, arguments);
      }

      return onSubmit;
    }()

    /**
     * On "subscribe" message.
     *
     * @param {Object} err
     * @param {Object} message
     */

  }, {
    key: 'onSubscribe',
    value: function () {
      var _ref8 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee8(err, message) {
        var snapshot;
        return regeneratorRuntime.wrap(function _callee8$(_context8) {
          while (1) {
            switch (_context8.prev = _context8.next) {
              case 0:
                if (!err) {
                  _context8.next = 2;
                  break;
                }

                return _context8.abrupt('return', this.rejectDeferred('subscribe', err));

              case 2:
                snapshot = message.snapshot;

                if (!snapshot) {
                  _context8.next = 6;
                  break;
                }

                _context8.next = 6;
                return this.update(snapshot);

              case 6:
                this.resolveDeferred('subscribe', this.value);
                if (snapshot) this.emit('subscribe', this.value);

              case 8:
              case 'end':
                return _context8.stop();
            }
          }
        }, _callee8, this);
      }));

      function onSubscribe(_x14, _x15) {
        return _ref8.apply(this, arguments);
      }

      return onSubscribe;
    }()

    /**
     * On "unsubscribe" message.
     *
     * @param {Object} err
     */

  }, {
    key: 'onUnsubscribe',
    value: function () {
      var _ref9 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee9(err) {
        return regeneratorRuntime.wrap(function _callee9$(_context9) {
          while (1) {
            switch (_context9.prev = _context9.next) {
              case 0:
                if (!err) {
                  _context9.next = 2;
                  break;
                }

                return _context9.abrupt('return', this.rejectDeferred('unsubscribe', err));

              case 2:

                this.resolveDeferred('unsubscribe');
                this.emit('unsubscribe');

              case 4:
              case 'end':
                return _context9.stop();
            }
          }
        }, _callee9, this);
      }));

      function onUnsubscribe(_x16) {
        return _ref9.apply(this, arguments);
      }

      return onUnsubscribe;
    }()

    /**
     * On client disconnect.
     */

  }]);

  return Document;
}(_events2.default);

/**
 * Export.
 *
 * @type {Document}
 */

var _initialiseProps = function _initialiseProps() {
  var _this4 = this;

  this.onClientDisconnect = function () {
    var inflight = _this4.inflight,
        deferreds = _this4.deferreds,
        client = _this4.client,
        queue = _this4.queue,
        subscribed = _this4.subscribed;

    var hasPendingSubscribe = !!deferreds.subscribe;
    var hasPendingFetch = !!deferreds.fetch;

    // If there's a pending inflight operation, return it to the queue.
    if (inflight) {
      queue.unshift(inflight);
      _this4.inflight = null;
    }

    _this4.resolveDeferred('unsubscribe');

    // When the client reconnects, force re-send the messages, so that any
    // original deferreds can be resolved/rejected again.
    client.once('connect', function () {
      _this4.flush({ force: true });
      if (hasPendingSubscribe || subscribed) _this4.subscribe({ force: true });
      if (hasPendingFetch) _this4.fetch({ force: true });
    });
  };
};

exports.default = Document;