'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; };

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

var _debug = require('debug');

var _debug2 = _interopRequireDefault(_debug);

var _events = require('events');

var _events2 = _interopRequireDefault(_events);

var _clone = require('lodash/clone');

var _clone2 = _interopRequireDefault(_clone);

var _isEmpty = require('is-empty');

var _isEmpty2 = _interopRequireDefault(_isEmpty);

var _stream = require('stream');

var _client = require('../client/client');

var _client2 = _interopRequireDefault(_client);

var _connection = require('./connection');

var _connection2 = _interopRequireDefault(_connection);

var _nexusError = require('../utils/nexus-error');

var _nexusError2 = _interopRequireDefault(_nexusError);

var _pubsub = require('../adapters/pubsub');

var _pubsub2 = _interopRequireDefault(_pubsub);

var _store = require('../adapters/store');

var _store2 = _interopRequireDefault(_store);

var _transport = require('../adapters/transport');

var _transport2 = _interopRequireDefault(_transport);

var _checkers = require('../utils/checkers');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _asyncToGenerator(fn) { return function () { var gen = fn.apply(this, arguments); return new Promise(function (resolve, reject) { function step(key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { return Promise.resolve(value).then(function (value) { step("next", value); }, function (err) { step("throw", err); }); } } return step("next"); }); }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _possibleConstructorReturn(self, call) { if (!self) { throw new ReferenceError("this hasn't been initialised - super() hasn't been called"); } return call && (typeof call === "object" || typeof call === "function") ? call : self; }

function _inherits(subClass, superClass) { if (typeof superClass !== "function" && superClass !== null) { throw new TypeError("Super expression must either be null or a function, not " + typeof superClass); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, enumerable: false, writable: true, configurable: true } }); if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass; }

/**
 * Debug.
 *
 * @type {Function}
 */

var debug = (0, _debug2.default)('nexus:server');

/**
 * An auto-incrementing ID for subscriptions.
 *
 * @type {Number}
 */

var subscriptionUid = 0;

/**
 * Nexus server.
 *
 * @type {Server}
 */

var Server = function (_Emitter) {
  _inherits(Server, _Emitter);

  /**
   * Constructor.
   *
   * @param {Object} config
   */

  function Server() {
    var config = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};

    _classCallCheck(this, Server);

    var _this = _possibleConstructorReturn(this, (Server.__proto__ || Object.getPrototypeOf(Server)).call(this));

    var _config$store = config.store,
        store = _config$store === undefined ? new _store2.default() : _config$store,
        _config$pubsub = config.pubsub,
        pubsub = _config$pubsub === undefined ? new _pubsub2.default() : _config$pubsub,
        _config$types = config.types,
        types = _config$types === undefined ? [] : _config$types,
        _config$maxRetries = config.maxRetries,
        maxRetries = _config$maxRetries === undefined ? 100 : _config$maxRetries;


    (0, _checkers.checkPubsub)(pubsub);
    (0, _checkers.checkStore)(store);
    (0, _checkers.checkTypes)(types);

    _this.store = store;
    _this.types = types;
    _this.pubsub = pubsub;
    _this.maxRetries = maxRetries;
    _this.connections = {};
    _this.subscriptions = {};
    _this.subscribed = {};
    _this.closed = true;

    pubsub.on('publish', function (id, operation) {
      if (typeof id !== 'string') {
        throw new _nexusError2.default('pubsub_id_invalid', { value: id });
      }

      if ((typeof operation === 'undefined' ? 'undefined' : _typeof(operation)) !== 'object') {
        throw new _nexusError2.default('pubsub_operation_invalid', { value: operation });
      }

      _this.broadcast(id, operation);
    });
    return _this;
  }

  /**
   * Open the server's adapters.
   */

  _createClass(Server, [{
    key: 'open',
    value: function () {
      var _ref = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee() {
        var store, pubsub;
        return regeneratorRuntime.wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                debug('open');

                store = this.store, pubsub = this.pubsub;
                _context.next = 4;
                return Promise.all([store.open && store.open(), pubsub.open && pubsub.open()]);

              case 4:

                this.closed = false;

              case 5:
              case 'end':
                return _context.stop();
            }
          }
        }, _callee, this);
      }));

      function open() {
        return _ref.apply(this, arguments);
      }

      return open;
    }()

    /**
     * Close the server's adapters and any open connections.
     */

  }, {
    key: 'close',
    value: function () {
      var _ref2 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee2() {
        var store, pubsub, connections, id, connection;
        return regeneratorRuntime.wrap(function _callee2$(_context2) {
          while (1) {
            switch (_context2.prev = _context2.next) {
              case 0:
                debug('close');

                store = this.store, pubsub = this.pubsub, connections = this.connections;


                for (id in connections) {
                  connection = connections[id];

                  connection.close();
                }

                _context2.next = 5;
                return Promise.all([store.close && store.close(), pubsub.close && pubsub.close()]);

              case 5:

                this.closed = true;

              case 6:
              case 'end':
                return _context2.stop();
            }
          }
        }, _callee2, this);
      }));

      function close() {
        return _ref2.apply(this, arguments);
      }

      return close;
    }()

    /**
     * Attach the Nexus server to an HTTP `server`, so it will open when the
     * server begins listening for connections.
     *
     * @param {Http.Server} server
     */

  }, {
    key: 'attach',
    value: function attach(server) {
      var _this2 = this;

      server.on('listen', _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee3() {
        return regeneratorRuntime.wrap(function _callee3$(_context3) {
          while (1) {
            switch (_context3.prev = _context3.next) {
              case 0:
                _context3.prev = 0;
                _context3.next = 3;
                return _this2.open();

              case 3:
                _context3.next = 8;
                break;

              case 5:
                _context3.prev = 5;
                _context3.t0 = _context3['catch'](0);

                _this2.emit('error', _context3.t0);

              case 8:
              case 'end':
                return _context3.stop();
            }
          }
        }, _callee3, _this2, [[0, 5]]);
      })));

      server.on('close', _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee4() {
        return regeneratorRuntime.wrap(function _callee4$(_context4) {
          while (1) {
            switch (_context4.prev = _context4.next) {
              case 0:
                _context4.prev = 0;
                _context4.next = 3;
                return _this2.close();

              case 3:
                _context4.next = 8;
                break;

              case 5:
                _context4.prev = 5;
                _context4.t0 = _context4['catch'](0);

                _this2.emit('error', _context4.t0);

              case 8:
              case 'end':
                return _context4.stop();
            }
          }
        }, _callee4, _this2, [[0, 5]]);
      })));
    }

    /**
     * Initialize a new connection for a `transport` stream with `context`.
     *
     * @param {Stream} transport
     * @param {Object} context
     * @return {Connection}
     */

  }, {
    key: 'connect',
    value: function () {
      var _ref5 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee5(transport) {
        var context = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
        var closed, connections, server, connection, onError, id;
        return regeneratorRuntime.wrap(function _callee5$(_context5) {
          while (1) {
            switch (_context5.prev = _context5.next) {
              case 0:
                debug('listen');

                (0, _checkers.checkTransport)(transport);

                closed = this.closed, connections = this.connections;
                server = this;

                if (!closed) {
                  _context5.next = 6;
                  break;
                }

                throw new _nexusError2.default('server_closed');

              case 6:
                connection = new _connection2.default({ server: server, transport: transport, context: context });
                onError = this.emit.bind(this, 'error');
                id = connection.id;


                connections[id] = connection;

                connection.on('error', onError);
                connection.once('close', function () {
                  connection.removeListener('error', onError);
                  delete connections[id];
                });

              case 12:
              case 'end':
                return _context5.stop();
            }
          }
        }, _callee5, this);
      }));

      function connect(_x2) {
        return _ref5.apply(this, arguments);
      }

      return connect;
    }()

    /**
     * Create an in-memory client for the server with `context`.
     *
     * @param {Object} context
     * @return {Client}
     */

  }, {
    key: 'client',
    value: function () {
      var _ref6 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee6(context) {
        var types, transport, client;
        return regeneratorRuntime.wrap(function _callee6$(_context6) {
          while (1) {
            switch (_context6.prev = _context6.next) {
              case 0:
                types = this.types;
                transport = new _transport2.default();
                _context6.next = 4;
                return this.connect(transport.server, context);

              case 4:
                client = new _client2.default({ types: types, transport: transport.client });
                return _context6.abrupt('return', client);

              case 6:
              case 'end':
                return _context6.stop();
            }
          }
        }, _callee6, this);
      }));

      function client(_x4) {
        return _ref6.apply(this, arguments);
      }

      return client;
    }()

    /**
     * Submit an `operation` to a document by `id`.
     *
     * @param {String} id
     * @param {Object} operation
     * @param {Object} options
     * @return {Array}
     */

  }, {
    key: 'submit',
    value: function () {
      var _ref7 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee7(id, operation) {
        var options = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};

        var maxRetries, _options$retry, retry, context, snapshot, type, operations, existing, _iteratorNormalCompletion, _didIteratorError, _iteratorError, _iterator, _step, ex, success;

        return regeneratorRuntime.wrap(function _callee7$(_context7) {
          while (1) {
            switch (_context7.prev = _context7.next) {
              case 0:
                debug('submit', { id: id, operation: operation, options: options });

                (0, _checkers.checkOperation)(operation);

                maxRetries = this.maxRetries;
                _options$retry = options.retry, retry = _options$retry === undefined ? 0 : _options$retry, context = options.context;
                _context7.next = 6;
                return this.findSnapshot(id, { context: context });

              case 6:
                snapshot = _context7.sent;


                (0, _checkers.checkType)(snapshot.type, this.types);

                type = this.types.find(function (t) {
                  return t.name === snapshot.type;
                });
                operations = [];

                // The operation version is always from a previous snapshot version, so it
                // should never be larger than the snapshot.

                if (!(operation.version > snapshot.version)) {
                  _context7.next = 12;
                  break;
                }

                throw new _nexusError2.default('operation_version_mismatch', { operation: operation });

              case 12:
                if (!(operation.version < snapshot.version)) {
                  _context7.next = 48;
                  break;
                }

                _context7.next = 15;
                return this.listOperations(id, operation.version, snapshot.version, { context: context });

              case 15:
                existing = _context7.sent;
                _iteratorNormalCompletion = true;
                _didIteratorError = false;
                _iteratorError = undefined;
                _context7.prev = 19;
                _iterator = existing[Symbol.iterator]();

              case 21:
                if (_iteratorNormalCompletion = (_step = _iterator.next()).done) {
                  _context7.next = 34;
                  break;
                }

                ex = _step.value;

                if (!(operation.id === ex.id)) {
                  _context7.next = 25;
                  break;
                }

                throw new _nexusError2.default('operation_duplicate', { operation: operation });

              case 25:
                if (!(operation.version !== ex.version)) {
                  _context7.next = 27;
                  break;
                }

                throw new _nexusError2.default('transform_version_mismatch', { operation: operation, snapshot: snapshot });

              case 27:

                debug('transform', { id: id, operation: operation, existing: ex });

                // Transform the operation by the existing one, and bump its version.
                operation.data = type.transform(operation.data, ex.data, 'left');
                operation.version++;
                operations.push(ex);

              case 31:
                _iteratorNormalCompletion = true;
                _context7.next = 21;
                break;

              case 34:
                _context7.next = 40;
                break;

              case 36:
                _context7.prev = 36;
                _context7.t0 = _context7['catch'](19);
                _didIteratorError = true;
                _iteratorError = _context7.t0;

              case 40:
                _context7.prev = 40;
                _context7.prev = 41;

                if (!_iteratorNormalCompletion && _iterator.return) {
                  _iterator.return();
                }

              case 43:
                _context7.prev = 43;

                if (!_didIteratorError) {
                  _context7.next = 46;
                  break;
                }

                throw _iteratorError;

              case 46:
                return _context7.finish(43);

              case 47:
                return _context7.finish(40);

              case 48:
                if (!(operation.version !== snapshot.version)) {
                  _context7.next = 50;
                  break;
                }

                throw new _nexusError2.default('apply_version_mismatch', { operation: operation, snapshot: snapshot });

              case 50:

                debug('apply', { id: id, operation: operation, snapshot: snapshot });

                // Apply the operation to the snapshot, and bump the snapshot's version.
                snapshot.data = type.apply(snapshot.data, operation.data);
                snapshot.version++;

                // Attempt to commit the resolve operation and snapshot.
                _context7.next = 55;
                return this.commit(id, operation, snapshot, { context: context });

              case 55:
                success = _context7.sent;

                if (success) {
                  _context7.next = 63;
                  break;
                }

                if (!(retry <= maxRetries)) {
                  _context7.next = 62;
                  break;
                }

                debug('retry', { id: id, operation: operation, options: options });
                _context7.next = 61;
                return this.submit(id, operation, _extends({}, options, { retry: retry + 1 }));

              case 61:
                return _context7.abrupt('return', _context7.sent);

              case 62:
                throw new _nexusError2.default('operation_submit_failed', { operation: operation });

              case 63:

                this.publish(id, operation, { context: context });
                return _context7.abrupt('return', operations);

              case 65:
              case 'end':
                return _context7.stop();
            }
          }
        }, _callee7, this, [[19, 36, 40, 48], [41,, 43, 47]]);
      }));

      function submit(_x5, _x6) {
        return _ref7.apply(this, arguments);
      }

      return submit;
    }()

    /**
     * Commit an `operation` and `snapshot` to the store for a document by `id`,
     * returning a boolean of whether the commit attempt was successful.
     *
     * @param {String} id
     * @param {Object} operation
     * @param {Object} snapshot
     * @param {Object} options
     * @return {Boolean}
     */

  }, {
    key: 'commit',
    value: function () {
      var _ref8 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee8(id, operation, snapshot) {
        var options = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : {};
        var context, success;
        return regeneratorRuntime.wrap(function _callee8$(_context8) {
          while (1) {
            switch (_context8.prev = _context8.next) {
              case 0:
                debug('commit', { id: id, operation: operation, snapshot: snapshot, options: options });

                (0, _checkers.checkOperation)(operation);
                (0, _checkers.checkSnapshot)(snapshot);

                context = options.context;
                _context8.next = 6;
                return this.store.commit(id, operation, snapshot, { context: context });

              case 6:
                success = _context8.sent;

                if (!(typeof success !== 'boolean')) {
                  _context8.next = 9;
                  break;
                }

                throw new _nexusError2.default('store_commit_invalid');

              case 9:
                return _context8.abrupt('return', success);

              case 10:
              case 'end':
                return _context8.stop();
            }
          }
        }, _callee8, this);
      }));

      function commit(_x8, _x9, _x10) {
        return _ref8.apply(this, arguments);
      }

      return commit;
    }()

    /**
     * Find the latest snapshot of a document by `id`.
     *
     * @param {String} id
     * @return {Object}
     */

  }, {
    key: 'findSnapshot',
    value: function () {
      var _ref9 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee9(id) {
        var options = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
        var context, snapshot;
        return regeneratorRuntime.wrap(function _callee9$(_context9) {
          while (1) {
            switch (_context9.prev = _context9.next) {
              case 0:
                debug('findSnapshot', { id: id, options: options });

                context = options.context;
                _context9.next = 4;
                return this.store.findSnapshot(id, { context: context });

              case 4:
                snapshot = _context9.sent;


                (0, _checkers.checkSnapshot)(snapshot);
                return _context9.abrupt('return', snapshot);

              case 7:
              case 'end':
                return _context9.stop();
            }
          }
        }, _callee9, this);
      }));

      function findSnapshot(_x12) {
        return _ref9.apply(this, arguments);
      }

      return findSnapshot;
    }()

    /**
     * List all of the operations for a document by `id`, starting `from` version
     * and going up `to` version (non-inclusive).
     *
     * @param {String} id
     * @param {Number} from
     * @param {Number} to
     * @param {Object} options
     * @return {Array}
     */

  }, {
    key: 'listOperations',
    value: function () {
      var _ref10 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee10(id, from) {
        var to = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : Infinity;
        var options = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : {};
        var context, ops;
        return regeneratorRuntime.wrap(function _callee10$(_context10) {
          while (1) {
            switch (_context10.prev = _context10.next) {
              case 0:
                debug('listOperations', { id: id, from: from, to: to, options: options });

                context = options.context;
                _context10.next = 4;
                return this.store.listOperations(id, from, to, { context: context });

              case 4:
                ops = _context10.sent;

                ops.forEach(function (op) {
                  return (0, _checkers.checkOperation)(op);
                });
                return _context10.abrupt('return', ops);

              case 7:
              case 'end':
                return _context10.stop();
            }
          }
        }, _callee10, this);
      }));

      function listOperations(_x14, _x15) {
        return _ref10.apply(this, arguments);
      }

      return listOperations;
    }()

    /**
     * Subscribe to a document by `id` at `version`, broadcasting all future
     * operations published to the document to the subscription stream.
     *
     * @param {String} id
     * @param {Number} version
     * @param {Object} options
     * @return {Readable}
     */

  }, {
    key: 'subscribe',
    value: function () {
      var _ref11 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee12(id, version) {
        var _this3 = this;

        var options = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};
        var context, subs, subscription, uid;
        return regeneratorRuntime.wrap(function _callee12$(_context12) {
          while (1) {
            switch (_context12.prev = _context12.next) {
              case 0:
                debug('subscribe', { id: id, version: version });

                context = options.context;
                subs = this.subscriptions[id] || (this.subscriptions[id] = {});
                subscription = new _stream.Readable({ objectMode: true });
                uid = subscriptionUid++;


                subs[uid] = subscription;
                subscription._read = function () {};
                subscription.once('close', _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee11() {
                  var s;
                  return regeneratorRuntime.wrap(function _callee11$(_context11) {
                    while (1) {
                      switch (_context11.prev = _context11.next) {
                        case 0:
                          s = _this3.subscriptions[id];

                          if (s) {
                            _context11.next = 3;
                            break;
                          }

                          return _context11.abrupt('return');

                        case 3:
                          delete s[uid];

                          if (!(0, _isEmpty2.default)(s)) {
                            _context11.next = 7;
                            break;
                          }

                          _context11.next = 7;
                          return _this3.unsubscribe(id, { context: context });

                        case 7:
                        case 'end':
                          return _context11.stop();
                      }
                    }
                  }, _callee11, _this3);
                })));

                if (this.subscribed[id]) {
                  _context12.next = 12;
                  break;
                }

                _context12.next = 11;
                return this.pubsub.subscribe(id, { context: context });

              case 11:
                this.subscribed[id] = true;

              case 12:
                return _context12.abrupt('return', subscription);

              case 13:
              case 'end':
                return _context12.stop();
            }
          }
        }, _callee12, this);
      }));

      function subscribe(_x18, _x19) {
        return _ref11.apply(this, arguments);
      }

      return subscribe;
    }()

    /**
     * Unsubscribe from a document by `id`.
     *
     * @param {String} id
     * @param {Object} options
     */

  }, {
    key: 'unsubscribe',
    value: function () {
      var _ref13 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee13(id) {
        var options = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
        var context;
        return regeneratorRuntime.wrap(function _callee13$(_context13) {
          while (1) {
            switch (_context13.prev = _context13.next) {
              case 0:
                debug('unsubscribe', { id: id });

                context = options.context;

                delete this.subscriptions[id];
                delete this.subscribed[id];
                _context13.next = 6;
                return this.pubsub.unsubscribe(id, { context: context });

              case 6:
              case 'end':
                return _context13.stop();
            }
          }
        }, _callee13, this);
      }));

      function unsubscribe(_x21) {
        return _ref13.apply(this, arguments);
      }

      return unsubscribe;
    }()

    /**
     * Publish an `operation` to a document by `id`.
     *
     * @param {String} id
     * @param {Object} operation
     * @param {Object} options
     */

  }, {
    key: 'publish',
    value: function () {
      var _ref14 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee14(id, operation) {
        var options = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};
        var context;
        return regeneratorRuntime.wrap(function _callee14$(_context14) {
          while (1) {
            switch (_context14.prev = _context14.next) {
              case 0:
                debug('publish', { id: id, operation: operation });

                context = options.context;
                _context14.next = 4;
                return this.pubsub.publish(id, operation, { context: context });

              case 4:
              case 'end':
                return _context14.stop();
            }
          }
        }, _callee14, this);
      }));

      function publish(_x23, _x24) {
        return _ref14.apply(this, arguments);
      }

      return publish;
    }()

    /**
     * Broadcast an `operation` to all subscriptions on a document by `id`.
     *
     * @param {String} id
     * @param {Object} operation
     */

  }, {
    key: 'broadcast',
    value: function () {
      var _ref15 = _asyncToGenerator( /*#__PURE__*/regeneratorRuntime.mark(function _callee15(id, operation) {
        var subs, key, sub, copy;
        return regeneratorRuntime.wrap(function _callee15$(_context15) {
          while (1) {
            switch (_context15.prev = _context15.next) {
              case 0:
                debug('broadcast', { id: id, operation: operation });

                subs = this.subscriptions[id];

                if (subs) {
                  _context15.next = 4;
                  break;
                }

                return _context15.abrupt('return');

              case 4:

                for (key in subs) {
                  sub = subs[key];
                  copy = (0, _clone2.default)(operation);

                  sub.push(copy);
                }

              case 5:
              case 'end':
                return _context15.stop();
            }
          }
        }, _callee15, this);
      }));

      function broadcast(_x26, _x27) {
        return _ref15.apply(this, arguments);
      }

      return broadcast;
    }()
  }]);

  return Server;
}(_events2.default);

/**
 * Export.
 *
 * @type {Server}
 */

exports.default = Server;